import { useEffect, useMemo, useSyncExternalStore } from 'react'

import { getConfig } from '../config'
import { createProxy } from '../helpers/createProxy'
import { matchMedia } from '../helpers/matchMedia'
import {
  MediaQueries,
  MediaQueryKey,
  MediaQueryObject,
  MediaQueryState,
  TamaguiInternalConfig,
} from '../types'
import { useSafeRef } from './useSafeRef'

export const mediaState: MediaQueryState =
  // development only safeguard
  process.env.NODE_ENV === 'development'
    ? createProxy(
        {},
        {
          get(target, key) {
            if (
              typeof key === 'string' &&
              key[0] === '$' &&
              // dont error on $$typeof
              key[1] !== '$'
            ) {
              throw new Error(`Access mediaState should not use "$": ${key}`)
            }
            return Reflect.get(target, key)
          },
        }
      )
    : ({} as any)

export const mediaQueryConfig: MediaQueries = {}
export const getMedia = () => mediaState

// for SSR capture it at time of startup
let initState: MediaQueryState
export const getInitialMediaState = () => {
  return (getConfig().disableSSR ? mediaState : initState) || {}
}

let mediaKeysOrdered: string[]
export const getMediaKeyImportance = (key: string) => {
  if (process.env.NODE_ENV === 'development' && key[0] === '$') {
    throw new Error(`use short key`)
  }
  // + 2 because we set base usedKeys=1 in getSplitStyles and all media go above 1
  return mediaKeysOrdered.indexOf(key) + 2
}

const dispose = new Set<Function>()

export const configureMedia = (config: TamaguiInternalConfig) => {
  const { media, mediaQueryDefaultActive } = config
  if (!media) return
  for (const key in media) {
    mediaState[key] = mediaQueryDefaultActive?.[key] || false
  }
  Object.assign(mediaQueryConfig, media)
  initState = { ...mediaState }
  updateCurrentState()
  mediaKeysOrdered = Object.keys(media)
  if (config.disableSSR) {
    setupMediaListeners()
  }
}

function unlisten() {
  dispose.forEach((cb) => cb())
  dispose.clear()
}

/**
 * Note: This should *not* set the state on the first render!
 * Because to avoid hydration issues SSR must match the server
 * *and then* re-render with the actual media query state.
 */
let configuredKey = ''
function setupMediaListeners() {
  // avoid setting up more than once per config
  const nextKey = JSON.stringify(mediaQueryConfig)
  if (nextKey === configuredKey) return
  configuredKey = nextKey

  // hmr, undo existing before re-binding
  unlisten()

  for (const key in mediaQueryConfig) {
    const str = mediaObjectToString(mediaQueryConfig[key])
    const getMatch = () => matchMedia(str)
    const match = getMatch()
    if (!match) {
      throw new Error('⚠️ No match')
    }
    // react native needs these deprecated apis for now
    match.addListener(update)
    dispose.add(() => match.removeListener(update))

    update()

    function update() {
      const next = !!getMatch().matches
      if (next === mediaState[key]) return
      mediaState[key] = next
      updateCurrentState()
    }
  }
}

export function useMediaListeners(config: TamaguiInternalConfig) {
  if (config.disableSSR) return
  useEffect(() => {
    setupMediaListeners()
    return unlisten
  }, [])
}

const currentStateListeners = new Set<any>()
function updateCurrentState() {
  currentStateListeners.forEach((cb) => cb(mediaState))
}
function subscribe(subscriber: any) {
  currentStateListeners.add(subscriber)
  return () => currentStateListeners.delete(subscriber)
}

type MediaKeysState = {
  [key: string]: any
}

type UseMediaState = {
  [key in MediaQueryKey]: boolean
}

const shouldUpdate = new WeakMap<any, boolean>()
export function setMediaShouldUpdate(ref: any, val: boolean) {
  return shouldUpdate.set(ref, val)
}

type UseMediaInternalState = { prev: MediaKeysState; next: MediaKeysState }
const empty = {}
let initialUseState: UseMediaInternalState

export function useMedia(): UseMediaState {
  // setup once
  initialUseState ||= {
    prev: initState,
    next: empty,
  }

  const keys = useSafeRef<UseMediaInternalState>(initialUseState)
  const state = useSyncExternalStore<MediaQueryState>(
    subscribe,
    () => {
      const curState = keys.current
      const preventUpdate = arguments[0] && shouldUpdate.get(arguments[0]) !== true
      if (preventUpdate) {
        return curState.prev
      }

      let didUpdate = false

      // ensure we keep the minimal object of only used keys
      const next = {}

      for (const key in curState.next) {
        const val = mediaState[key]
        if (curState.prev[key] != val) {
          didUpdate = true
        }
        next[key] = val
      }

      // console.log('go', { didUpdate, next, mediaState, curState })

      if (!didUpdate) {
        return curState.prev
      }

      keys.current = { next, prev: next }
      return next
    },
    () => initState
  )

  return useMemo(() => {
    return new Proxy(state === empty ? initState : state, {
      get(_, key: string) {
        if (!(key in keys.current.next)) {
          keys.current.next[key] = true
          keys.current = { ...keys.current }
        }
        return Reflect.get(state, key)
      },
    })
  }, [state])
}

/**
 * Useful for more complex components that need access to the currently active props,
 * accounting for the currently active media queries.
 *
 * Use sparingly, is will loop props and trigger re-render on all media queries.
 *
 * */
export function useMediaPropsActive<A extends Object>(
  props: A
): {
  // remove all media
  [Key in keyof A extends `$${string}` ? never : keyof A]?: A[Key]
} {
  const media = useMedia()

  return useMemo(() => {
    const next = {} as A
    const importancesUsed = {}
    const propNames = Object.keys(props)

    for (let i = propNames.length - 1; i >= 0; i--) {
      const key = propNames[i]
      const val = props[key]
      if (key[0] === '$') {
        const mediaKey = key.slice(1)
        if (!media[mediaKey]) continue
        if (val && typeof val === 'object') {
          const subKeys = Object.keys(val)
          for (let j = subKeys.length; j--; j >= 0) {
            const subKey = subKeys[j]
            mergeMediaByImportance(next, mediaKey, subKey, val[subKey], importancesUsed)
          }
        }
      } else {
        mergeMediaByImportance(next, '', key, val, importancesUsed)
      }
    }

    return next
  }, [media, props])
}

export const getMediaImportanceIfMoreImportant = (
  mediaKey: string,
  key: string,
  importancesUsed: Record<string, number>
) => {
  const importance = getMediaKeyImportance(mediaKey)
  return !importancesUsed[key] || importance > importancesUsed[key] ? importance : null
}

export function mergeMediaByImportance(
  onto: Record<string, any>,
  mediaKey: string,
  key: string,
  value: any,
  importancesUsed: Record<string, number>
) {
  const importance = getMediaImportanceIfMoreImportant(mediaKey, key, importancesUsed)
  if (importance === null) {
    return false
  }
  importancesUsed[key] = importance
  onto[key] = value
  return true
}

function camelToHyphen(str: string) {
  return str.replace(/[A-Z]/g, (m) => `-${m.toLowerCase()}`).toLowerCase()
}

export function mediaObjectToString(query: string | MediaQueryObject) {
  if (typeof query === 'string') {
    return query
  }
  return Object.entries(query)
    .map(([feature, value]) => {
      feature = camelToHyphen(feature)
      if (typeof value === 'string') {
        return `(${feature}: ${value})`
      }
      if (typeof value === 'number' && /[height|width]$/.test(feature)) {
        value = `${value}px`
      }
      return `(${feature}: ${value})`
    })
    .join(' and ')
}
